/*************************************************************************

Copyright (c) 2018 Novell, Inc.  All Rights Reserved.

With respect to this file, Novell hereby grants to Developer a
royalty-free, non-exclusive license to include this sample code
and derivative binaries in its product. Novell grants to Developer
worldwide distribution rights to market, distribute or sell this
sample code file and derivative binaries as a component of
Developer's product(s).  Novell shall have no obligations to
Developer or Developer's customers with respect to this code.

DISCLAIMER:

Novell disclaims and excludes any and all express, implied, and
statutory warranties, including, without limitation, warranties
of good title, warranties against infringement, and the implied
warranties of merchantibility and fitness for a particular purpose.
Novell does not warrant that the software will satisfy customer's
requirements or that the licensed works are without defect or error
or that the operation of the software will be uninterrupted.
Novell makes no warranties respecting any technical services or
support tools provided under the agreement, and disclaims all other
warranties, including the implied warranties of merchantability and
fitness for a particular purpose. */

/**************************************************************************
NWRecoverDeletedFileExt.C
**************************************************************************/


#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <dirent.h>
#include <ctype.h>
#include <unistd.h>


#include <nwdel.h>
#include <nwcalls.h>
#include <nwclxcon.h>
#include <nwnet.h>
#include <nwerror.h>

#define MAX_NO_OF_ENTRIES 100 //For sample application we are assuming max of entries to be 100, but this value can vary in different enviroment.

void NWTimeToLinuxTime(nuint32 dateTime, char * pszOut, int cchOutSize);
void DisplayData(NWCONN_HANDLE hConnection, NWDSContextHandle hContextHandle, nuint16 NoOfEntries, NWDELETED_INFO_EXT2 deletedInfo []);
NWRCODE eDirectoryLoginIfNeeded(NWCONN_HANDLE * phConnection, NWDSContextHandle * phContextHandle, char * pszServerName, char * pszUserObject, char * pszPassword, nbool * pbNewLoginPerformed);

int main(int argc, char **argv)
{
	char szNWPath[1024] = { 0 }, szUserObject[MAX_DN_CHARS] = { 0 }, szPassword[32] = { 0 }, szServerName[NW_MAX_SERVER_NAME_LEN] = { 0 };

	NWRCODE ccode;

	NWDSContextHandle hContextHandle = (NWDSContextHandle) -1;
	NWCONN_HANDLE hConnection = 0;
	nbool bNewLoginPerformed = FALSE;

	NWDIR_HANDLE hDirectory = 0;
	NWCONN_HANDLE hNewConnection = 0;
	NWSCANFORDELETED_HANDLE hScanHandle = 0;

	// Separate our first line of output from the command line, whatever that first line ends up being.
	printf("\n");

	// Initialize our ability to call the SDK functions.
	ccode = NWCallsInit(NULL, NULL);

	if (ccode != SUCCESS)
	{
		printf("NWCallsInit failed with code 0x%04X (%d).\n", ccode, ccode);
		return ccode; // Skip all clean-up; we never got off the ground.
	}

	if (argc != 5)
	{
		printf("Usage:\n  recoverDeletedFile <server> <userObject> <password> <vol:path\\path>\n");
		goto CleanUpExit;
	}
	else	
	{
		strcpy(szServerName, argv[1]);
		strcpy(szUserObject, argv[2]);
		strcpy(szPassword, argv[3]);
		strcpy(szNWPath, argv[4]);
	}


	// Test whether we are already logged in to the eDirectory tree implied
	// by the server name.  If we're not already logged on, use the provided
	// user object DN and password to attempt logging in.

	ccode = eDirectoryLoginIfNeeded(&hConnection, &hContextHandle, szServerName /* server */, szUserObject /* user */, szPassword /* password */, &bNewLoginPerformed);

	if (ccode != SUCCESS)
		goto CleanUpExit;

	// Open a directory handle to the path specified.  Continue with whatever connection
	// handle is returned, in case a junction has taken us to a different server.
	// Not only so our directory handle is relative to the connection handle we're
	// specifying, but also so we have correct connection handle for NWDSMapIDToName
	// for the eDirectory entry ID numbers this scan will return.

	ccode = NWAllocTempNSDirHandle3(hConnection, 0, szNWPath, NW_NS_LONG, &hNewConnection, &hDirectory, NW_NS_LONG);

	if (ccode != 0)
	{
		printf("NWAllocTempNSDirHandle3() for \"%s\" returned 0x%04X (%d).\n", szNWPath, ccode, ccode);
		goto CleanUpExit;
	}

	if (hNewConnection)
	{
		NWCCCloseConn(hConnection);
		hConnection = hNewConnection;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////Now using the XPLAT API NWScanForDeletedFilesExt2 to scan all the deleted files in the given directory path////////////////////////////////////////////

	nuint16 TotolNoOfEntries = 0;
	NWDELETED_INFO_EXT2 allDeletedInfo[MAX_NO_OF_ENTRIES];
	memset(allDeletedInfo, 0, sizeof(NWDELETED_INFO_EXT2) * MAX_NO_OF_ENTRIES);
	nuint32 volNum = 0;
	nuint32 dirBase = 0;

	do
	{
		//Don't Proceed if we have reached maximum limit
		if (TotolNoOfEntries >= MAX_NO_OF_ENTRIES)
		{
			printf("Error: OUT OF MEMORY!!! Maximum Limit of %d is reached for storing Deleted File(s) Information!!!!\n", MAX_NO_OF_ENTRIES);
			break;
		}

		NWDELETED_INFO_EXT2 deletedInfo[10] = { {0} };
		nuint16 NoOfEntries = sizeof(deletedInfo) / sizeof(deletedInfo[0]);

		ccode = NWScanForDeletedFilesExt2(&hScanHandle, hConnection, hDirectory, &volNum, &dirBase, &NoOfEntries, deletedInfo);

		if (ccode)
		{
			if (ccode == NO_MORE_ENTRIES)
			{
				printf("Finished enumerating the available deleted files.\n");
				break;
			}
			else
			{
				printf("NWScanForDeletedFilesExt2 failed with unexpected error 0x%04X (%d).\n", ccode, ccode);
				goto CleanUpExit;
			}
		}

		//Only display and store the deleted file(s) infomation till we reach maximum limit defined by MAX_NO_OF_ENTRIES
		if (TotolNoOfEntries + NoOfEntries >= MAX_NO_OF_ENTRIES)
			NoOfEntries = MAX_NO_OF_ENTRIES - TotolNoOfEntries;

		//Copy the entries to the allDeletedInfo
		int i; 
		for (i = 0; i < NoOfEntries; i++)
		{
			memcpy(&allDeletedInfo[TotolNoOfEntries++], &deletedInfo[i], sizeof(NWDELETED_INFO_EXT2));
		}

		// Display all the deleted files information 
		DisplayData(hConnection, hContextHandle, NoOfEntries, deletedInfo);

	} while (!ccode);


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////Now using the XPLAT API NWRecoverDeletedFileExt to recover deleted files in the given directory path//////////////////////////////////////////////

	if (TotolNoOfEntries == 0)
	{
		printf("\n\nNo Files are available for the path %s that can be recovered(or salvaged), hence exiting the application!!!!!!\n", szNWPath);
		goto CleanUpExit;
	}

	int input = 0;

	NWRECOVER_FILE_INFO allRecoverFileInfo[MAX_NO_OF_ENTRIES];
	nuint16 noOfRecoverFileEntries = 0;
	do
	{
		printf("\n\n*******************************************************************************************************************************************************************************************************************************");
		printf("\n*******************************************************************************************************************************************************************************************************************************\n");
		printf("\n\n");
		printf("\n\t\t\t\t\t\t SALVAGE MENU \n");
		printf("\n\t\t\t 1. Press 1 for Recover(Or Salvage) All Above Files\n");
		printf("\n\t\t\t 2. Press 2 to Recover(Or Salvage) selective files using Item No\n");
		printf("\n\t\t\t 3. Press 3 to exit\n");

		printf("\n\t\t\t Enter your choice(1-3) : ");
		scanf("%d", &input);

		switch (input)
		{
			case 1: {
					printf("\n*******************************************************************************************************************************************************************************************************************************\n");
					printf("\n Recovering(Salvaging) All file for directory %s\n ", szNWPath);
					printf("\n-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
					//Create the object of NWRECOVER_FILE_INFO using NWDELETED_INFO_EXT2
					memset(allRecoverFileInfo, 0, sizeof(NWRECOVER_FILE_INFO) * MAX_NO_OF_ENTRIES);
					int i;
					for (i = 0; i < TotolNoOfEntries; i++)
					{
						allRecoverFileInfo[i].fileSysFlag = allDeletedInfo[i].fileSysFlag;
						allRecoverFileInfo[i].sequence = allDeletedInfo[i].sequence;
						strcpy((char *)allRecoverFileInfo[i].rcvrFileName, (const char *)allDeletedInfo[i].name);
					}

					//Now call NWRecoverDeletedFileExt2 to recover the file(s)
					noOfRecoverFileEntries = TotolNoOfEntries;
					ccode = NWRecoverDeletedFileExt2(hConnection, hDirectory, volNum, dirBase, noOfRecoverFileEntries, allRecoverFileInfo);
					if (ccode)
					{
						printf("\n NWScanForDeletedFilesExt2 on connectionID 0x%08X failed. Status=0x%08X.\n", hConnection, ccode);
					}

					for (i = 0; i < noOfRecoverFileEntries; i++)
					{

						if (allRecoverFileInfo[i].errorcode == 0)
							printf("\n Recover(or Salvage) for file %s succeeded.\n", allRecoverFileInfo[i].rcvrFileName);
						else
							printf("\n Recover(or Salvage) for file %s Failed with error code =0x%08X.\n", allRecoverFileInfo[i].rcvrFileName, allRecoverFileInfo[i].errorcode);
					}
				}
				goto CleanUpExit;

			case 2: {
					char inputstr[100];
					char *token = NULL, *nextToken = NULL;
					int inputdata[MAX_NO_OF_ENTRIES], count = 0;
					printf("\n\t\t\t Enter the Item No or list of Item No(s)(serparated by space) from the above list of files to recover or salvage: ");
					// flushes the standard input (clears the input buffer)
					while ((getchar()) != '\n');
					fgets(inputstr, 100, stdin);
					token = strtok_r(inputstr, " ", &nextToken);
					while (token != NULL)
					{
						int sno = strtol(token, NULL, 10);
						if (sno != 0 && sno <= TotolNoOfEntries)
						{
							inputdata[count++] = sno;
						}
						else
							printf("\n\t\t\t %s is an invalid Item No!!!!", token);

						token = strtok_r(NULL, " ", &nextToken);
					}

					if (count == 0)
					{
						printf("\n\t\t\t Invalid Item No(s) entered, hence Exiting the application!!!!\n");
						goto CleanUpExit;
					}

					printf("\n*******************************************************************************************************************************************************************************************************************************\n");
					printf("\n Recovering(Salvaging) file(s) for directory %s\n ", szNWPath);
					printf("\n-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");

					//Create the object of NWRECOVER_FILE_INFO using NWDELETED_INFO_EXT2
					memset(allRecoverFileInfo, 0, sizeof(NWRECOVER_FILE_INFO) * MAX_NO_OF_ENTRIES);
					int i;
					for (i = 0; i < count; i++)
					{
						int sno = inputdata[i];
						allRecoverFileInfo[i].fileSysFlag = allDeletedInfo[sno -1].fileSysFlag;
						allRecoverFileInfo[i].sequence = allDeletedInfo[sno -1].sequence;
						strcpy((char *)allRecoverFileInfo[i].rcvrFileName, (const char *)allDeletedInfo[sno -1].name);
					}

					//Now call NWRecoverDeletedFileExt2 to recover the file(s)
					noOfRecoverFileEntries = count;
					ccode = NWRecoverDeletedFileExt2(hConnection, hDirectory, volNum, dirBase, noOfRecoverFileEntries, allRecoverFileInfo);
					if (ccode)
					{
						printf("\n NWScanForDeletedFilesExt2 on connectionID 0x%08X failed. Status=0x%08X.\n", hConnection, ccode);
					}

					for (i = 0; i < noOfRecoverFileEntries; i++)
					{

						if (allRecoverFileInfo[i].errorcode == 0)
							printf("\n Recover(or Salvage) for file %s succeeded.\n", allRecoverFileInfo[i].rcvrFileName);
						else
							printf("\n Recover(or Salvage) for file %s Failed with error code =0x%08X.\n", allRecoverFileInfo[i].rcvrFileName, allRecoverFileInfo[i].errorcode);
					}
				}	
				goto CleanUpExit;

			case 3: printf("\n\t\t\t Exiting the Program...............................................\n");
				goto CleanUpExit;
				break;

			default: printf("\n\t\t\t Please Enter a valid Input(1 - 3)!!!!!!!!!!!!!!!!!\n");
		}

	} while (1);

CleanUpExit:

	if (hScanHandle)
		NWCloseScanForDeletedFilesHandle(&hScanHandle);

	if (bNewLoginPerformed)
		NWDSLogout(hContextHandle);

	if ((NWDSContextHandle) -1 != hContextHandle)
		NWDSFreeContext(hContextHandle);

	if (hConnection)
		NWCCCloseConn(hConnection);

	NWCallsTerm(NULL);

	return(ccode);
}

// Display the Data on the console
void DisplayData(NWCONN_HANDLE hConnection, NWDSContextHandle hContextHandle, nuint16 NoOfEntries, NWDELETED_INFO_EXT2 deletedInfo [])
{
	static int NumberOfDeletedFiles = 1;

	if (NumberOfDeletedFiles == 1)
		printf("\nAvailable deleted files:\n\n");

	int tempCnt;
	for ( tempCnt = 0; tempCnt < NoOfEntries; tempCnt++ )
	{
		char szDeletorName[MAX_DN_CHARS + 1] = "";
		char szOwnerName[MAX_DN_CHARS + 1] = "";
		char szCreationTime[MAX_DN_CHARS + 1] = "";
		char szDeletedTime[MAX_DN_CHARS + 1] = "";
		char szLastUpdatedTime[MAX_DN_CHARS + 1] = "";
		char szLastAccessTime[MAX_DN_CHARS + 1] = "";

		// Creation Date & Time
		NWTimeToLinuxTime(deletedInfo[tempCnt].creationDateAndTime, szCreationTime, sizeof(szCreationTime) / sizeof(szCreationTime[0]));

		// Deleted Date & Time
		NWTimeToLinuxTime(deletedInfo[tempCnt].deletedDateAndTime, szDeletedTime, sizeof(szDeletedTime) / sizeof(szDeletedTime[0]));

		// Last Updated Date & Time
		NWTimeToLinuxTime(deletedInfo[0].updateDateAndTime, szLastUpdatedTime, sizeof(szLastUpdatedTime) / sizeof(szLastUpdatedTime[0]));

		// Last Access Date (time is not provided, will always be midnight)
		NWTimeToLinuxTime(((nuint32) deletedInfo[tempCnt].lastAccessDate) << 16, szLastAccessTime, sizeof(szLastAccessTime) / sizeof(szLastAccessTime[0]));

		// Deletor Name
		NWDSMapIDToName(hContextHandle, hConnection, deletedInfo[tempCnt].deletorID, (pnstr8) szDeletorName);

		// Owner Name
		NWDSMapIDToName(hContextHandle, hConnection, deletedInfo[tempCnt].ownerID, (pnstr8) szOwnerName);

		// Display all data
		printf(
				"Item %u: \"%s\"\n"
				"     Deleted: %s  Deleted By: %s\n"
				"     File Size: %u KB  Last Modified: %s\n"
				"     Last Accessed: %s  Created: %s\n"
				"     Owner: %s\n"
				"\n",
				NumberOfDeletedFiles++, deletedInfo[tempCnt].name,
				szDeletedTime, szDeletorName,
				deletedInfo[tempCnt].fileSize / 1024, szLastUpdatedTime,
				szLastAccessTime, szCreationTime,
				szOwnerName);
	}

	return;
}

// Convert the date and time to locale-specific short display.
void NWTimeToLinuxTime(nuint32 dateTime, char * pszOut, int cchOutSize)
{
	NW_DATE date = { 0 };
	NW_TIME time = { 0 };

	// Ensure caller's buffer is NULL-terminated if we don't actually fill it.
	*pszOut = '\0';

	NWUnpackDateTime(dateTime, &date, &time);

	if (date.day && date.month) // Can't be zero, if actual date present.
	{	
		sprintf(pszOut, "%02d-%02d-%04d %02d:%02d:%02d",date.month, date.day, date.year, time.hours, time.minutes, time.seconds);
	}

	return;
}


// Determine whether we're using an existing login or performing a new eDirectory login.
NWRCODE eDirectoryLoginIfNeeded
(
 NWCONN_HANDLE * phConnection,
 NWDSContextHandle * phContextHandle,
 char * pszServerName,
 char * pszUserObject,
 char * pszPassword,
 nbool * pbNewLoginPerformed
)
{
	NWRCODE ccode;
	char szTreeName[NW_MAX_TREE_NAME_LEN] = { 0 }; // NULL-init entire buffer, to make removal of padding easier.

	ccode = NWCCOpenConnByName(0, pszServerName, NWCC_NAME_FORMAT_BIND, NWCC_OPEN_LICENSED, NWCC_TRAN_TYPE_WILD, phConnection);

	if ( ccode == SUCCESS )
	{
		printf("Opened connection to server \"%s\".\n", pszServerName);

		ccode = NWCCGetConnInfo(*phConnection, NWCC_INFO_TREE_NAME, sizeof( szTreeName ), szTreeName);

		if (ccode == SUCCESS)
		{
			// Tree name returned by connection is padded with underscores.
			// Strip by walking backwards through the buffer, so that we
			// don't remove underscores that are part of the tree name itself.

			int i;
			for ( i = sizeof( szTreeName ) - 1; i > 0; i-- )
			{
				if (szTreeName[i] && (szTreeName[i] != '_'))
					break; // Have reached the end of actual tree name.

				if (szTreeName[i] == '_')
					szTreeName[i] = '\0'; // Replce underscore that is not part of tree name with a NULL.
			}

			printf( "Tree name: \"%s\".\n", szTreeName );

			ccode = NWDSCreateContextHandle(phContextHandle);

			if (ccode == SUCCESS)
			{
				// There must be a tree name in the context handle or a default tree name

				NWDSSetContext(*phContextHandle, DCK_TREE_NAME, szTreeName);

				if (NWDSCanDSAuthenticate(*phContextHandle) != TRUE)
				{

					NWDSSetContext( *phContextHandle , DCK_NAME_CONTEXT, "" );

					ccode = NWDSLoginEx( *phContextHandle, (pnstr8) pszUserObject, PWD_RAW_C_STRING, pszPassword );

					if ( ccode == SUCCESS )
					{
						printf( "Logged into eDirectory successfully as \"%s\".\n", pszUserObject );

						if ( pbNewLoginPerformed )
							*pbNewLoginPerformed = TRUE;
					}
					else if ( ccode == NWE_ALREADY_ATTACHED )
					{
						// We're logged in already, but why didn't NWDSCanDSAuthenticate already know this?
						// Still, if NWDSAuthenticateConn can succeed, its valid to proceed for this sample.

						printf( "NWDSLoginEx() returned NWE_ALREADY_ATTACHED 0x%04X (%d).\n", ccode, ccode );
					}
					else
						printf( "NWDSLoginEx() as \"%s\" returned 0x%04X (%d).\n", pszUserObject, ccode, ccode );

				}
				else
				{
					// Looks like an eDirectory login already exists for the specified tree / server.
					// So don't call NWDSLoginEx to create a new login, as only 1 connection is allowed
					// to a given tree/server per linux user session.  We'll use the existing login to
					// perform the required action.

					printf( "NWDSCanDSAuthenticate indicated eDirectory credentials already exist for tree \"%s\". Hence, not attempting a new login.\n", szTreeName );
				}

				// Regardless of whether we had to perform a new eDirectory login or not,
				// ensure that the NCP connection the caller provided us can be authenticated,
				// since it may or may not have already been inherently involved or authenticated
				// during the eDirectory login processing.

				ccode = NWDSAuthenticateConn(*phContextHandle, *phConnection);

				if (ccode)
					printf("NWDSAuthenticateConn() returned 0x%04X (%d).\n", ccode, ccode);
			}
			else
				printf("NWDSCreateContextHandle() returned 0x%04X (%d).\n", ccode, ccode);
		}
		else
			printf("NWCCGetConnInfo() returned 0x%04X (%d).\n", ccode, ccode);
	}
	else
		printf("NWCCOpenConnByName() returned 0x%04X (%d).\n", ccode, ccode);

	return ccode;
}

